import torch
import triton
import triton.language as tl


def paged_attention_v1(
    query: torch.Tensor,        # [num_seqs, num_heads, head_dim]
    key_cache: torch.Tensor,    # [num_blocks, num_kv_heads, block_size, head_dim]
    value_cache: torch.Tensor,  # [num_blocks, num_kv_heads, block_size, head_dim]
    block_tables: torch.Tensor, # [num_seqs, max_num_blocks_per_seq]
    context_lens: torch.Tensor, # [num_seqs]
    scale: float
) -> torch.Tensor:              # [num_seqs, num_heads, head_dim]
    
    num_seqs, num_heads, head_dim = query.shape
    _, num_kv_heads, block_size, _ = key_cache.shape
    out = torch.zeros(num_seqs, num_heads, head_dim, device=query.device, dtype=query.dtype)

    for i in range(num_seqs):
        context_len = context_lens[i]
        num_blocks = (context_len + block_size - 1) // block_size
        
        # block_idx: [num_used_blocks]
        block_idx = block_tables[i, :num_blocks]

        # keys/values: [num_used_blocks, num_kv_heads, block_size, head_dim]
        keys = key_cache[block_idx]      
        values = value_cache[block_idx]  
        
        # [num_blocks, heads, block_size, dim] -> [num_blocks * block_size, heads, dim]
        keys = keys.transpose(1, 2).reshape(-1, num_kv_heads, head_dim)
        values = values.transpose(1, 2).reshape(-1, num_kv_heads, head_dim)

        # keys/values: [context_len, num_kv_heads, head_dim]
        keys = keys[:context_len]
        values = values[:context_len]

        # q: [num_heads, 1, head_dim]
        q = query[i].unsqueeze(1)
        
        # k: [num_kv_heads, head_dim, context_len]
        k = keys.transpose(0, 1).transpose(1, 2)
        # v: [num_kv_heads, context_len, head_dim]
        v = values.transpose(0, 1)

        if num_kv_heads != num_heads: # For GQA
            # k/v: [num_heads, head_dim, context_len]
            k = torch.repeat_interleave(k, num_heads // num_kv_heads, dim=0)
            v = torch.repeat_interleave(v, num_heads // num_kv_heads, dim=0)

        # score: [num_heads, 1, context_len]
        score = torch.matmul(q, k) * scale
        attn = torch.softmax(score, dim=-1)

        # out[i]: [num_heads, head_dim]
        out[i] = torch.matmul(attn, v).squeeze(1)

    return out

@triton.jit
def _paged_attention_kernel(
    q_ptr,      # [num_seqs, num_heads, head_dim]
    k_ptr,      # [num_blocks, num_kv_heads, block_size, head_dim]
    v_ptr,      # [num_blocks, num_kv_heads, block_size, head_dim]
    block_ptr,  # [num_seqs, max_num_blocks]
    lens_ptr,   # [num_seqs]
    out_ptr,    # [num_seqs, num_heads, head_dim]
    stride_q_batch, stride_q_head, stride_q_dim,
    stride_k_batch, stride_k_head, stride_k_dim, stride_k_x,
    stride_v_batch, stride_v_head, stride_v_dim, stride_v_x,
    stride_out_batch, stride_out_head, stride_out_dim,
    stride_b_batch, stride_b_block,
    num_heads,
    num_kv_heads,
    scale: tl.constexpr,
    block_size: tl.constexpr,
    head_dim: tl.constexpr,
):
    pid_batch = tl.program_id(0)
    pid_head = tl.program_id(1)

    offs_dim = tl.arange(0, head_dim)
    q_ptr_offset = (
        pid_batch * stride_q_batch + 
        pid_head * stride_q_head +
        offs_dim * stride_q_dim
    )
    q = tl.load(q_ptr + q_ptr_offset) # load query vector into SRAM

    m_i = -float("inf") # max score so far
    l_i = 0.0           # denominator of softmax
    acc = tl.zeros([head_dim], dtype=tl.float32)

    context_len = tl.load(lens_ptr + pid_batch)
    num_blocks = (context_len + block_size - 1) // block_size

    for block_idx in range(num_blocks):
        block_table_ptr = block_ptr + pid_batch * stride_b_batch + block_idx * stride_b_block
        physical_block_id = tl.load(block_table_ptr)

        offs_block = tl.arange(0, block_size)

        kv_head_idx = pid_head // (num_heads // num_kv_heads)

        k_ptr_base = k_ptr + physical_block_id * stride_k_batch + kv_head_idx * stride_k_head
        v_ptr_base = v_ptr + physical_block_id * stride_v_batch + kv_head_idx * stride_v_head

        k_offsets = offs_block[:, None] * stride_k_x + offs_dim[None, :] * stride_k_dim
        v_offsets = offs_block[:, None] * stride_v_x + offs_dim[None, :] * stride_v_dim

        k = tl.load(k_ptr_base + k_offsets)
        v = tl.load(v_ptr_base + v_offsets)

        score = tl.sum(q[None, :] * k, axis=1)
        score *= scale

        token_indices = block_idx * block_size + offs_block
        mask = token_indices < context_len
        score = tl.where(mask, score, -float("inf"))

        m_prev = m_i
        m_i = tl.maximum(m_i, tl.max(score, axis=0))

        p = tl.exp(score - m_i)

        l_i = l_i * tl.exp(m_prev - m_i) + tl.sum(p, axis=0)

        acc = acc * tl.exp(m_prev - m_i) + tl.sum(p[:, None] * v, axis=0)

    acc = acc / l_i

    out_ptr_offset = (
        pid_batch * stride_out_batch + 
        pid_head * stride_out_head +
        offs_dim * stride_out_dim
    )
    tl.store(out_ptr + out_ptr_offset, acc.to(q.dtype))

def paged_attention_triton(query, key_cache, value_cache, block_tables, context_lens, scale):
    num_seqs, num_heads, head_dim = query.shape
    num_blocks, num_kv_heads, block_size, _ = key_cache.shape

    output = torch.empty_like(query)

    grid = (num_seqs, num_heads)

    _paged_attention_kernel[grid](
        query, key_cache, value_cache, block_tables, context_lens,
        output,
        query.stride(0), query.stride(1), query.stride(2),
        key_cache.stride(0), key_cache.stride(1), key_cache.stride(3), key_cache.stride(2),
        value_cache.stride(0), value_cache.stride(1), value_cache.stride(3), value_cache.stride(2),
        output.stride(0), output.stride(1), output.stride(2),
        block_tables.stride(0), block_tables.stride(1),
        num_heads, num_kv_heads, scale, block_size, head_dim
    )

    return output



import torch
from transformers import AutoTokenizer

from core.block_manager import BlockAllocator, BlockSpaceManager
from core.model import ModelExecutor
from core.scheduler import Scheduler
from core.sequence import Sequence, SequenceGroup


class LLMEngine:
    def __init__(self, model_name: str, block_size: int=16, num_gpu_blocks: int=100):
        self.block_size = block_size
        self.allocator = BlockAllocator(num_blocks=num_gpu_blocks, block_size=block_size)
        self.block_manager = BlockSpaceManager(self.allocator)
        self.scheduler = Scheduler(self.block_manager)

        self.model_executor = ModelExecutor(model_name, num_gpu_blocks, block_size)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.eos_token_id = self.tokenizer.eos_token_id

        self.request_counter = 0
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def add_request(self, prompt: str) -> str:
        req_id = str(self.request_counter)
        self.request_counter += 1
        seq_id = int(req_id)

        token_ids = self.tokenizer.encode(prompt)

        seq = Sequence(seq_id, prompt, token_ids)
        group = SequenceGroup(req_id, [seq], arrival_time=0)

        self.scheduler.add_sequence_group(group)
        return req_id

    @torch.inference_mode()
    def step(self):
        running_groups = self.scheduler.schedule()
        if not running_groups:
            return {}

        first_seq = running_groups[0].get_seqs()[0]
        is_prefill = len(first_seq.output_token_ids) == 0

        input_ids_list = []
        position_ids_list = []
        context_lens_list = []
        block_tables_list = []

        for group in running_groups:
            seq = group.get_seqs()[0]
            
            if is_prefill:
                input_ids_list.append(seq.get_token_ids())
                position_ids_list.append(list(range(seq.get_len())))
            else:
                input_ids_list.append(seq.get_token_ids()[-1])
                position_ids_list.append(seq.get_len() - 1)
            context_lens_list.append(seq.get_len())
            block_tables_list.append(self.block_manager.get_block_table(seq.seq_id))

        max_num_blocks = max([len(block_table) for block_table in block_tables_list])
        pad_id = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id is not None else 0

        if is_prefill:
            # input_tensor: [batch_size, max_seq_len]
            input_tensor = torch.nn.utils.rnn.pad_sequence(
                [torch.tensor(ids) for ids in input_ids_list],
                batch_first=True,
                padding_value=pad_id
            ).to(self.device)
            # position_tensor: [batch_size, max_seq_len]
            position_tensor = torch.nn.utils.rnn.pad_sequence(
                [torch.tensor(ids) for ids in position_ids_list],
                batch_first=True,
                padding_value=pad_id
            ).to(self.device)
        else:
            # input_tensor: [batch_size, 1]
            input_tensor = torch.tensor(input_ids_list, device=self.device, dtype=torch.long).unsqueeze(1) 
            # position_tensor: [batch_size, 1]
            position_tensor = torch.tensor(position_ids_list, device=self.device, dtype=torch.long).unsqueeze(1)

        padded_block_tables = []
        for blocks in block_tables_list:
            num_pad = max_num_blocks - len(blocks)
            padded_block_tables.append(blocks + [-1] * num_pad)

        # context_lens_tensor: [batch_size]
        context_lens_tensor = torch.tensor(context_lens_list, device=self.device, dtype=torch.int64)   
        # block_tables_tensor: [batch_size, max_num_blocks]
        block_tables_tensor = torch.tensor(padded_block_tables, device=self.device, dtype=torch.int64) 

        # logits: [batch_size, seq_len, vocab_size]
        logits = self.model_executor.forward(input_tensor, position_tensor, context_lens_tensor, block_tables_tensor, is_prefill)

        outputs = {}
        for i, group in enumerate(running_groups):
            seq = group.get_seqs()[0]
    
            # logit_idx = (context_lens_list[i] - 1) if is_prefill else 0
            next_token_logits = logits[i, -1, :] # Shape: [vocab_size]
            # next_token_id = torch.argmax(next_token_logits).item()
            
            temperature = 0.8
            #
            probs = torch.softmax(next_token_logits / temperature, dim=-1)
            next_token_id = torch.multinomial(probs, num_samples=1).item()
            #
            # top_p = 0.9
            # sorted_probs, sorted_indices = torch.sort(probs, descending=True)
            # cum_probs = torch.cumsum(sorted_probs, dim=-1)
            # sorted_indices_to_remove = cum_probs > top_p
            # sorted_indices_to_remove[0] = False # Keep the top token
            # sorted_probs[sorted_indices_to_remove] = 0
            # sorted_probs = sorted_probs / sorted_probs.sum()
            #
            # sampled_idx = torch.multinomial(sorted_probs, num_samples=1).item()
            # next_token_id = sorted_indices[sampled_idx].item()

            seq.append_token_id(next_token_id, 1.0)

            from core.sequence import SequenceStatus

            if next_token_id == self.eos_token_id:
                seq.status = SequenceStatus.FINISHED
                self.block_manager.free(seq.seq_id)
            else:
                self.block_manager.append_slot(seq.seq_id, seq.get_len() - 1)

            outputs[group.request_id] = self.tokenizer.decode(seq.get_token_ids())

        return outputs



from unittest.mock import patch

import torch

from core.llm_engine import LLMEngine


# Mock the heavy dependencies
@patch("core.llm_engine.AutoTokenizer")
@patch("core.llm_engine.ModelExecutor")
def test_engine_step(mock_model_cls, mock_tokenizer_cls):
    # 1. Setup Mocks
    # Mock Tokenizer
    mock_tokenizer = mock_tokenizer_cls.from_pretrained.return_value
    mock_tokenizer.encode.return_value = [1, 2, 3] # "Hello world"
    mock_tokenizer.pad_token_id = 0
    mock_tokenizer.decode.return_value = "Hello world generated"
    
    # Mock Model
    mock_model_instance = mock_model_cls.return_value
    # Return random logits: [batch=1, seq_len=4, vocab=100]
    # We need to ensure the shape matches what step() expects
    def mock_forward(input_ids):
        batch, seq_len = input_ids.shape
        return torch.randn(batch, seq_len, 100) 
    
    mock_model_instance.forward.side_effect = mock_forward

    # 2. Initialize Engine
    # Use block_size=2 to test allocation easily
    engine = LLMEngine("dummy-model", block_size=2, num_gpu_blocks=10)
    
    # 3. Add Request
    req_id = engine.add_request("Hello world")
    
    # 4. Step 1
    outputs = engine.step()
    
    # 5. Verify
    assert req_id in outputs
    # Sequence length should increase from 3 to 4
    seq = engine.scheduler.running[0].get_seqs()[0]
    assert seq.get_len() == 4
    # Block manager should have been called
    # (3 tokens -> 2 blocks. 4 tokens -> 2 blocks. No new allocation yet).
    assert len(engine.block_manager.block_tables[int(req_id)]) == 2
    
    print("Test passed!")


from collections import deque
from typing import Deque, List

from core.block_manager import BlockSpaceManager
from core.sequence import SequenceGroup


class Scheduler:
    def __init__(self, block_manager: BlockSpaceManager):
        self.block_manager = block_manager
        self.waiting: Deque[SequenceGroup] = deque()
        self.running: List[SequenceGroup] = []
        self.swapped: List[SequenceGroup] = []

    def add_sequence_group(self, seq_group: SequenceGroup):
        self.waiting.append(seq_group)

    def schedule(self) -> List[SequenceGroup]:
        self.running = [g for g in self.running if not g.is_finished()]

        while self.waiting:
            seq_group = self.waiting[0]
            seq = seq_group.get_seqs()[0]

            token_ids = seq.get_token_ids()

            if self.block_manager.can_allocate_request(token_ids):
                self.waiting.popleft()
                self.block_manager.allocate_request(seq.seq_id, token_ids)
                self.running.append(seq_group)
            else:
                break

        return self.running



import torch
from transformers import AutoModelForCausalLM

from kernels.attention import paged_attention_triton


def rotate_half(x: torch.Tensor):
    x1 = x[..., :x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2:]
    return torch.cat((-x2, x1), dim=-1)

def apply_rotary_pos_emb(q, k, cos, sin, position_ids=None, unsqueeze_dim=1):
    cos = cos.unsqueeze(unsqueeze_dim)
    sin = sin.unsqueeze(unsqueeze_dim)

    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)

    return q_embed, k_embed

class ModelExecutor:
    def __init__(self, model_name: str, num_gpu_blocks: int, block_size: int):
        self.num_gpu_blocks = num_gpu_blocks
        self.block_size = block_size
        self.model = AutoModelForCausalLM.from_pretrained(
            model_name,
            dtype=torch.bfloat16,
            device_map="cuda"
        )

        self.num_heads = getattr(self.model.config, "num_attention_heads", getattr(self.model.config, "n_head", None))
        self.num_layers = getattr(self.model.config, "num_hidden_layers", getattr(self.model.config, "n_layer", None))
        self.hidden_size = getattr(self.model.config, "hidden_size", getattr(self.model.config, "n_embd", None))

        head_dim = getattr(self.model.config, "head_dim", None)
        if head_dim is None:
            self.head_dim = self.hidden_size // self.num_heads
        else:
            self.head_dim = head_dim

        self.num_kv_heads = getattr(self.model.config, "num_key_value_heads", self.num_heads)

        self._init_kv_cache()

        if hasattr(self.model, "model"):
            self.layers = self.model.model.layers
        elif hasattr(self.model, "transformer"):
            self.layers = self.model.transformer.h
        else:
            raise ValueError("Could not find layers attribute in model")

        for i, layer in enumerate(self.layers):
            attn_module = None
            if hasattr(self.layers, "self_attn"):
                attn_module = layer.self_attn
            elif hasattr(self.layers, "attn"):
                attn_module = layer.attn

            if attn_module is None:
                if "self_attn" in layer._modules:
                    attn_module = layer._modules["self_attn"]
                elif "attn" in layer._modules:
                    attn_module = layer._modules["attn"]

            if not attn_module:
                raise ValueError(f"Unknown attention module. Available modules: {layer._modules.keys()}")

            attn_module.layer_idx = i
            object.__setattr__(attn_module, "forward", self._make_forward(attn_module))

    def _init_kv_cache(self):
        self.kv_cache = [torch.zeros(
            2,
            self.num_gpu_blocks,
            self.num_kv_heads,
            self.block_size,
            self.head_dim,
            device=self.model.device,
            dtype=self.model.dtype
        ) for _ in range(self.num_layers)]

    def _make_forward(self, module):
        def forward_wrapper(hidden_states, *args, **kwargs):
            return self._paged_attention_forward(module, hidden_states, *args, **kwargs)
        return forward_wrapper

    def forward(self, input_ids: torch.Tensor, position_ids: torch.Tensor, context_lens: torch.Tensor, block_tables: torch.Tensor, is_prefill: bool=False):
        self.context_lens = context_lens
        self.block_tables = block_tables
        self.is_prefill = is_prefill

        outputs = self.model(input_ids=input_ids, position_ids=position_ids)
        return outputs.logits

    def _paged_attention_forward(self, module, hidden_states: torch.Tensor, *args, **kwargs):
        # hidden_states: [batch_size, seq_len, hidden_size]
        if hasattr(module, "c_attn"):
            # GPT2 style
            qkv = module.c_attn(hidden_states)
            query, key, value = qkv.split(self.head_dim * self.num_heads, dim=2)
        else:
            # Llama style
            query = module.q_proj(hidden_states)
            key = module.k_proj(hidden_states)
            value = module.v_proj(hidden_states)

        # Q: [batch_size, seq_len, num_heads * head_dim]
        # K: [batch_size, seq_len, num_kv_heads * head_dim]
        batch_size, seq_len, _ = hidden_states.shape
        executor = self

        # query: [batch_size, num_heads, seq_len, head_dim]
        # key:   [batch_size, num_kv_heads, seq_len, head_dim]
        query = query.view(batch_size, seq_len, executor.num_heads, executor.head_dim).transpose(1, 2)
        key = key.view(batch_size, seq_len, executor.num_kv_heads, executor.head_dim).transpose(1, 2)
        value = value.view(batch_size, seq_len, executor.num_kv_heads, executor.head_dim).transpose(1, 2)

        rotary_emb = getattr(self.model.model, "rotary_emb", None)
        if rotary_emb is not None:
            pos_ids = kwargs.get("position_ids")
            if pos_ids is not None:
                # cos, sin: [batch_size, seq_len, head_dim]
                cos, sin = rotary_emb(value, pos_ids)
                query, key = apply_rotary_pos_emb(query, key, cos, sin)
            else:
                pos_emb = kwargs.get("position_embeddings")
                if pos_emb is not None:
                    cos, sin = pos_emb
                    query, key = apply_rotary_pos_emb(query, key, cos, sin)

        layer_idx = module.layer_idx
        # k_cache: [num_blocks, num_kv_heads, block_size, head_dim]
        k_cache = executor.kv_cache[layer_idx][0]
        v_cache = executor.kv_cache[layer_idx][1]

        if executor.is_prefill:
            # PREFILL PHASE: Write prompt to cache
            if executor.num_kv_heads != executor.num_heads:
                key_expanded = torch.repeat_interleave(key, executor.num_heads // executor.num_kv_heads, dim=1)
                value_expanded = torch.repeat_interleave(value, executor.num_heads // executor.num_kv_heads, dim=1)
                attn_output = torch.nn.functional.scaled_dot_product_attention(query, key_expanded, value_expanded, is_causal=True)
            else:
                attn_output = torch.nn.functional.scaled_dot_product_attention(query, key, value, is_causal=True)

            for i in range(batch_size):
                length = executor.context_lens[i]
                block_table = executor.block_tables[i]

                # k_seq: [seq_len, num_kv_heads, head_dim]
                k_seq = key[i, :, :length, :].permute(1, 0, 2)
                v_seq = value[i, :, :length, :].permute(1, 0, 2)

                for t in range(length):
                    block_indices = block_table[t // executor.block_size]
                    block_offsets = t % executor.block_size
                    
                    k_cache[block_indices, :, block_offsets, :] = k_seq[t]
                    v_cache[block_indices, :, block_offsets, :] = v_seq[t]
            
            # attn_output: [batch_size, heads, seq_len, head_dim]
            attn_output = attn_output.transpose(1, 2).contiguous()
            if hasattr(module, "c_proj"):
                output = module.c_proj(attn_output.view(batch_size * seq_len, -1))
            else:
                output = module.o_proj(attn_output.view(batch_size * seq_len, -1))
            return output.view(batch_size, seq_len, -1), None
        else:
            # DECODE PHASE: Read from cache
            # query: [batch_size, num_heads, head_dim]
            query = query.squeeze(2)
            key = key.squeeze(2)
            value = value.squeeze(2)

            last_token_indices = executor.context_lens - 1
            block_indices = executor.block_tables.gather(1, (last_token_indices // executor.block_size).unsqueeze(1).long()).squeeze(1)
            block_offsets = last_token_indices % executor.block_size

            # k_cache[block_idx, :, offset, :] = key
            k_cache[block_indices, :, block_offsets, :] = key
            v_cache[block_indices, :, block_offsets, :] = value

            scale = 1.0 / (executor.head_dim ** 0.5)
            
            # attn_output = paged_attention_v1(query, k_cache, v_cache, executor.block_tables, executor.context_lens, scale)
            attn_output = paged_attention_triton(query, k_cache, v_cache, executor.block_tables, executor.context_lens, scale)

            if hasattr(module, "c_proj"):
                output = module.c_proj(attn_output.view(batch_size, -1))
            else:
                output = module.o_proj(attn_output.view(batch_size, -1))
            return output.view(batch_size, 1, -1), None



from unittest.mock import patch
import torch
from core.llm_engine import LLMEngine

@patch("core.llm_engine.AutoTokenizer")
@patch("core.llm_engine.ModelExecutor")
def test_mixed_batch_bug(mock_model_cls, mock_tokenizer_cls):
    # Setup Mocks
    mock_tokenizer = mock_tokenizer_cls.from_pretrained.return_value
    mock_tokenizer.encode.side_effect = lambda x: [1, 2, 3] if "A" in x else [4, 5, 6]
    mock_tokenizer.pad_token_id = 0
    mock_tokenizer.decode.return_value = "generated"

    # Mock Model Instance
    mock_model_instance = mock_model_cls.return_value
    
    # We want to capture the inputs to forward
    forward_inputs = []
    def mock_forward(input_ids, context_lens, block_tables, is_prefill):
        forward_inputs.append({
            "input_ids": input_ids,
            "is_prefill": is_prefill
        })
        batch_size = input_ids.shape[0]
        seq_len = input_ids.shape[1]
        # Return dummy logits with correct shape
        return torch.randn(batch_size, seq_len, 100) 
        
    mock_model_instance.forward.side_effect = mock_forward

    engine = LLMEngine("dummy", block_size=16, num_gpu_blocks=10)

    # 1. Add Request A
    engine.add_request("Prompt A") # ids: [1, 2, 3]

    # 2. Step 1 (Prefill A)
    engine.step()
    
    print("Step 1 done")
    assert forward_inputs[-1]["is_prefill"]
    assert forward_inputs[-1]["input_ids"].shape == (1, 3) # Batch 1, Seq 3

    # 3. Add Request B
    engine.add_request("Prompt B") # ids: [4, 5, 6]

    # 4. Step 2 (Decode A + Prefill B)
    # This is where the bug should manifest
    engine.step()
    
    print("Step 2 done")
    last_call = forward_inputs[-1]
    
    # The engine takes is_prefill from the first group.
    # If Scheduler puts A first, is_prefill will be False.
    # If Scheduler puts B first, is_prefill will be True.
    # Scheduler uses deque and appends to running. 
    # A is already in running. B is added to running. So A is first.
    
    is_prefill_arg = last_call["is_prefill"]
    input_ids_arg = last_call["input_ids"]
    
    print(f"Is Prefill: {is_prefill_arg}")
    print(f"Input Shape: {input_ids_arg.shape}")
    print(f"Input IDs: {input_ids_arg}")

    # Expectation: 
    # is_prefill should be False (because A is first).
    # Request B should have been processed as prefill (3 tokens), but treated as decode (1 token).
    
    # Check Request B inputs (index 1)
    # B's tokens are [4, 5, 6]. 
    # If treated as decode, we only see [6].
    
    # input_ids_arg shape will be [2, 1] because is_prefill=False forces stack with unsqueeze(1) or similar logic in engine
    
    if not is_prefill_arg:
        # We expected mixed batch handling, but if engine forces False, B is broken.
        b_input = input_ids_arg[1]
        if len(b_input) == 1 and b_input[0] == 6:
            print("BUG CONFIRMED: Request B (prefill) was processed as decode (last token only).")
        else:
            print("Something else happened.")
    else:
        print("Unexpected: is_prefill was True. Did A finish?")

if __name__ == "__main__":
    test_mixed_batch_bug()



import pytest
from core.scheduler import BlockAllocator, PhysicalTokenBlock

def test_allocator_initialization():
    allocator = BlockAllocator(num_blocks=10, block_size=16)
    assert allocator.get_num_free_blocks() == 10
    
def test_allocate_single_block():
    allocator = BlockAllocator(num_blocks=10, block_size=16)
    block = allocator.allocate()
    
    assert isinstance(block, PhysicalTokenBlock)
    assert block.block_num == 0
    assert block.ref_count == 1
    assert allocator.get_num_free_blocks() == 9

def test_allocate_all_blocks():
    allocator = BlockAllocator(num_blocks=2, block_size=16)
    allocator.allocate()
    allocator.allocate()
    
    assert allocator.get_num_free_blocks() == 0
    
    # Next allocation should fail
    with pytest.raises(ValueError, match="Out of memory"):
        allocator.allocate()

def test_free_block():
    allocator = BlockAllocator(num_blocks=10, block_size=16)
    block = allocator.allocate()
    
    assert allocator.get_num_free_blocks() == 9
    
    allocator.free(block)
    assert block.ref_count == 0
    assert allocator.get_num_free_blocks() == 10
    
    # Re-allocating should give us the same block back (LIFO/FIFO depending on deque)
    # Your implementation uses popleft (FIFO) for allocate and append (end) for free
    # So it behaves like a queue.
    
def test_double_free_error():
    allocator = BlockAllocator(num_blocks=10, block_size=16)
    block = allocator.allocate()
    
    allocator.free(block)
    
    with pytest.raises(ValueError, match="already free"):
        allocator.free(block)


import torch
import unittest
from unittest.mock import MagicMock, patch
from core.llm_engine import LLMEngine

class MockConfig:
    num_attention_heads = 4
    num_hidden_layers = 1
    hidden_size = 16
    head_dim = 4
    num_key_value_heads = 4
    n_head = 4
    n_layer = 1
    n_embd = 16

class MockAttention(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.q_proj = torch.nn.Linear(16, 16, bias=False, dtype=torch.float16)
        self.k_proj = torch.nn.Linear(16, 16, bias=False, dtype=torch.float16)
        self.v_proj = torch.nn.Linear(16, 16, bias=False, dtype=torch.float16)
        self.o_proj = torch.nn.Linear(16, 16, bias=False, dtype=torch.float16)
        self.layer_idx = 0

class MockLayer(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.self_attn = MockAttention()
    
    def forward(self, hidden_states, *args, **kwargs):
        # We must call self_attn to trigger the patched method
        result = self.self_attn(hidden_states)
        if isinstance(result, tuple):
            return result[0]
        return result

class MockModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = torch.nn.ModuleList([MockLayer()])
        self.config = MockConfig()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
    
    def forward(self, input_ids):
        # Create dummy embeddings with correct seq_len
        batch_size = input_ids.shape[0]
        seq_len = input_ids.shape[1]
        hidden_states = torch.randn(batch_size, seq_len, 16, device=self.device, dtype=torch.float16)
        
        for layer in self.layers:
            hidden_states = layer(hidden_states)
            
        return hidden_states

class MockHFModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = MockModel()
        self.config = MockConfig()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
    
    def forward(self, input_ids):
        # Call the internal model
        hidden_states = self.model(input_ids)
        
        # Return logits as expected by ModelExecutor
        # Logits should have shape [batch, seq_len, vocab_size]
        batch_size, seq_len, _ = hidden_states.shape
        logits = torch.randn(batch_size, seq_len, 100, device=self.device, dtype=torch.float16)
        return MagicMock(logits=logits)

class TestPagedAttentionIntegration(unittest.TestCase):
    @patch("core.model.AutoModelForCausalLM")
    @patch("core.llm_engine.AutoTokenizer")
    def test_end_to_end_step(self, mock_tokenizer_cls, mock_model_cls):
        # 1. Setup HF Mocks
        device = "cuda" if torch.cuda.is_available() else "cpu"
        mock_hf_model = MockHFModel().to(device).to(torch.float16)
        mock_model_cls.from_pretrained.return_value = mock_hf_model
        
        mock_tokenizer = mock_tokenizer_cls.from_pretrained.return_value
        # Prompt: 3 tokens [1, 2, 3]
        mock_tokenizer.encode.return_value = [1, 2, 3]
        mock_tokenizer.pad_token_id = 0
        mock_tokenizer.decode.return_value = "generated"

        # 2. Init Engine (This triggers ModelExecutor init and PATCHING)
        # block_size=2, so 3 tokens = 2 blocks:
        # Block 0: indices 0, 1
        # Block 1: index 2
        engine = LLMEngine("dummy", block_size=2, num_gpu_blocks=10)
        
        # 3. Add Request (Prefill phase)
        engine.add_request("test prompt")
        
        # 4. Run Step 1 (Prefill)
        print("Running Prefill step...")
        with torch.no_grad():
            engine.step()
        
        # Verify Prefill cache writing
        seq_id = 0
        block_table = engine.block_manager.get_block_table(seq_id)
        kv_cache = engine.model_executor.kv_cache
        
        # Check Block 0 (Tokens 0 and 1)
        b0_idx = block_table[0]
        # Data at Block 0, offset 0 (Token 0)
        assert kv_cache[0][0][b0_idx, :, 0, :].abs().sum() > 0.0
        # Data at Block 0, offset 1 (Token 1)
        assert kv_cache[0][0][b0_idx, :, 1, :].abs().sum() > 0.0
        
        # Check Block 1 (Token 2)
        b1_idx = block_table[1]
        assert kv_cache[0][0][b1_idx, :, 0, :].abs().sum() > 0.0

        # 5. Run Step 2 (Decode)
        # Now context_len is 4 (3 prompt + 1 generated)
        print("Running Decode step...")
        with torch.no_grad():
            engine.step()
        
        # Check if 4th token (index 3) was written to Block 1, offset 1
        assert kv_cache[0][0][b1_idx, :, 1, :].abs().sum() > 0.0
        
        print("All assertions passed!")


if __name__ == "__main__":
    unittest.main()



import logging
import os
import time

import modal
from dotenv import load_dotenv

load_dotenv()

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S"
)
logger = logging.getLogger(__name__)

app = modal.App("mini-vllm")

image = (
    modal.Image.debian_slim()
    .pip_install(
        "torch",
        "triton",
        "transformers",
        "huggingface-hub",
        "einops",
        "numpy<2.0.0",
        "accelerate",
        "python-dotenv",
        "hf_transfer"
    )
    .env({"HF_HUB_ENABLE_HF_TRANSFER": "1"}) # to avoid model download every time
    .add_local_dir("core", remote_path="/root/core")
    .add_local_dir("kernels", remote_path="/root/kernels")
)

# MODEL_NAME = "google/gemma-3-270m"
# MODEL_NAME = "openai-community/gpt2"
MODEL_NAME = "meta-llama/Llama-3.2-1B"
GPU_CONFIG = "A100"

@app.cls(
    gpu=GPU_CONFIG,
    image=image,
    timeout=600,
    secrets=[modal.Secret.from_dict({"HF_TOKEN": os.getenv("HF_TOKEN")})],
)
class InferenceEngine:
    @modal.enter()
    def load_model(self):
        # import torch
        # from transformers import AutoModelForCausalLM, AutoTokenizer
        from transformers import AutoTokenizer

        from core.llm_engine import LLMEngine

        logger.info(f"Loading model {MODEL_NAME}...")
        t0 = time.time()

        self.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
        # self.model = AutoModelForCausalLM.from_pretrained(
        #     MODEL_NAME,
        #     dtype=torch.bfloat16,
        #     device_map="auto"
        # )
        self.engine = LLMEngine(MODEL_NAME, num_gpu_blocks=5000)

        logger.info(f"Model loaded in {time.time() - t0:.2f}s")

    @modal.method()
    def generate(self, prompt: str):
        import time
        t0= time.time()

        # inputs = self.tokenizer(prompt, return_tensors="pt").to("cuda")

        # outputs = self.model.generate(
        #     **inputs,
        #     max_new_tokens=20
        # )

        # generated_text = self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        req_id = self.engine.add_request(prompt)

        final_text = ""
        token_count = 0
        while True:
            outputs = self.engine.step()
            token_count += 1
            if req_id in outputs:
                final_text = outputs[req_id]
                pass
            else:
                if final_text:
                    break
                if not outputs:
                    break
            
            if token_count > 256:
                break

        duration = time.time() - t0

        return {
            "text": final_text,
            "duration_sec": duration
        }

    @modal.method()
    def generate_batch(self, prompts: list):
        import time
        t0 = time.time()

        req_ids = []
        for prompt in prompts:
            req_ids.append(self.engine.add_request(prompt))

        current_texts = {req_id: "" for req_id in req_ids}
        finished = {}
        token_count = 0

        while len(finished) < len(req_ids):
            outputs = self.engine.step()
            token_count += 1

            for req_id in req_ids:
                if req_id in outputs:
                    current_texts[req_id] = outputs[req_id]
                elif req_id not in finished and current_texts[req_id]:
                    finished[req_id] = current_texts[req_id]

            if token_count > 256:
                for req_id in req_ids:
                    if req_id not in finished:
                        finished[req_id] = current_texts[req_id]
                break

        return {
            "texts": finished,
            "duration_sec": time.time() - t0
        }

@app.local_entrypoint()
def main():
    # prompt = [
    #     "The meaning of life is to find",
    #     "The meaning of life is to give"
    # ]
    prompt = [
        "You are a helpful AI assistant. You always provide detailed, accurate, and thoughtful answers to questions. Please explain the concept of gravity.",
        "You are a helpful AI assistant. You always provide detailed, accurate, and thoughtful answers to questions. Please describe how plants grow.",
    ]
    logger.info(f"Sending prompt: '{prompt}'")

    engine = InferenceEngine()
    results = engine.generate_batch.remote(prompt)
    for req_id, text in results["texts"].items():
        logger.info(f"[{req_id}]: {text}")

    # logger.info("--- Result ---")
    # logger.info(f"Output: {result['text']}")
    # logger.info(f"Time: {result['duration_sec']:.2f}")



import os
import time

import modal
from dotenv import load_dotenv

load_dotenv()

app = modal.App("mini-vllm-benchmark")

# Image for mini-vllm
mini_vllm_image = (
    modal.Image.debian_slim()
    .pip_install(
        "torch",
        "triton",
        "transformers",
        "huggingface-hub",
        "einops",
        "numpy<2.0.0",
        "accelerate",
        "python-dotenv",
        "hf_transfer"
    )
    .env({"HF_HUB_ENABLE_HF_TRANSFER": "1"})
    .add_local_dir("core", remote_path="/root/core")
    .add_local_dir("kernels", remote_path="/root/kernels")
)

# Image for vLLM
vllm_image = (
    modal.Image.debian_slim()
    .pip_install(
        "vllm",
        "python-dotenv",
        "hf_transfer"
    )
    .env({"HF_HUB_ENABLE_HF_TRANSFER": "1"})
)

MODEL_NAME = "meta-llama/Llama-3.2-1B"
GPU_CONFIG = "A100"
PROMPT = "The meaning of life is"


@app.cls(
    gpu=GPU_CONFIG,
    image=mini_vllm_image,
    timeout=600,
    secrets=[modal.Secret.from_dict({"HF_TOKEN": os.getenv("HF_TOKEN")})],
)
class MiniVLLMBenchmark:
    @modal.enter()
    def load_model(self):
        from core.llm_engine import LLMEngine
        print(f"[mini-vllm] Loading model {MODEL_NAME}...")
        self.engine = LLMEngine(MODEL_NAME, num_gpu_blocks=5000)
        print("[mini-vllm] Model loaded.")

    @modal.method()
    def run_throughput_test(self, num_requests: int = 4, steps: int = 50):
        req_ids = []
        for i in range(num_requests):
            req_id = self.engine.add_request(PROMPT)
            req_ids.append(req_id)
        
        start_time = time.time()
        total_tokens_generated = 0
        
        for _ in range(steps):
            outputs = self.engine.step()
            total_tokens_generated += len(outputs)
            if not outputs:
                break

        end_time = time.time()
        duration = end_time - start_time
        
        return {
            "duration": duration,
            "total_tokens": total_tokens_generated,
            "tokens_per_sec": total_tokens_generated / duration,
            "num_requests": num_requests
        }


@app.cls(
    gpu=GPU_CONFIG,
    image=vllm_image,
    timeout=600,
    secrets=[modal.Secret.from_dict({"HF_TOKEN": os.getenv("HF_TOKEN")})],
)
class VLLMBenchmark:
    @modal.enter()
    def load_model(self):
        from vllm import LLM
        print(f"[vLLM] Loading model {MODEL_NAME}...")
        self.llm = LLM(model=MODEL_NAME, gpu_memory_utilization=0.9)
        print("[vLLM] Model loaded.")

    @modal.method()
    def run_throughput_test(self, num_requests: int = 4, max_tokens: int = 50):
        from vllm import SamplingParams
        
        # Create prompts for batch
        prompts = [PROMPT] * num_requests
        sampling_params = SamplingParams(
            temperature=0,  # Greedy to match mini-vllm
            max_tokens=max_tokens
        )
        
        start_time = time.time()
        outputs = self.llm.generate(prompts, sampling_params)
        end_time = time.time()
        
        duration = end_time - start_time
        total_tokens = sum(len(o.outputs[0].token_ids) for o in outputs)
        
        return {
            "duration": duration,
            "total_tokens": total_tokens,
            "tokens_per_sec": total_tokens / duration,
            "num_requests": num_requests
        }


@app.local_entrypoint()
def main():
    print("=" * 60)
    print("BENCHMARK: mini-vllm vs vLLM")
    print("=" * 60)
    
    scenarios = [
        {"requests": 1, "steps": 50},
        {"requests": 4, "steps": 50},
        {"requests": 16, "steps": 50},
    ]
    
    mini_results = []
    vllm_results = []
    
    # Run mini-vllm benchmarks
    print("\n" + "=" * 60)
    print("Running mini-vllm benchmarks...")
    print("=" * 60)
    mini_benchmark = MiniVLLMBenchmark()
    
    for s in scenarios:
        print(f"\n--- mini-vllm: Batch Size {s['requests']} ---")
        stats = mini_benchmark.run_throughput_test.remote(s['requests'], s['steps'])
        mini_results.append(stats)
        print(f"Duration: {stats['duration']:.4f}s")
        print(f"Total Tokens: {stats['total_tokens']}")
        print(f"Throughput: {stats['tokens_per_sec']:.2f} tokens/sec")
    
    # Run vLLM benchmarks
    print("\n" + "=" * 60)
    print("Running vLLM benchmarks...")
    print("=" * 60)
    vllm_benchmark = VLLMBenchmark()
    
    for s in scenarios:
        print(f"\n--- vLLM: Batch Size {s['requests']} ---")
        stats = vllm_benchmark.run_throughput_test.remote(s['requests'], s['steps'])
        vllm_results.append(stats)
        print(f"Duration: {stats['duration']:.4f}s")
        print(f"Total Tokens: {stats['total_tokens']}")
        print(f"Throughput: {stats['tokens_per_sec']:.2f} tokens/sec")
    
    # Print comparison table
    print("\n" + "=" * 60)
    print("COMPARISON SUMMARY")
    print("=" * 60)
    print(f"{'Batch':<8} {'mini-vllm':<18} {'vLLM':<18} {'Ratio':<10}")
    print(f"{'Size':<8} {'(tokens/sec)':<18} {'(tokens/sec)':<18} {'(vLLM/mini)':<10}")
    print("-" * 60)
    
    for i, s in enumerate(scenarios):
        mini_tps = mini_results[i]['tokens_per_sec']
        vllm_tps = vllm_results[i]['tokens_per_sec']
        ratio = vllm_tps / mini_tps if mini_tps > 0 else 0
        print(f"{s['requests']:<8} {mini_tps:<18.2f} {vllm_tps:<18.2f} {ratio:<10.2f}x")



from enum import Enum
from typing import List, Optional

class SequenceStatus(Enum):
    WAITING = 1
    RUNNING = 2
    FINISHED = 3

class Sequence:
    def __init__(
        self,
        seq_id: int,
        prompt: str,
        prompt_token_ids: List[int],
    ):
        self.seq_id = seq_id
        self.prompt = prompt
        self.prompt_token_ids = prompt_token_ids
        self.output_token_ids = []
        self.status = SequenceStatus.WAITING

    def append_token_id(self, token_id: int, logprob: float) -> None:
        self.output_token_ids.append(token_id)

    def get_len(self) -> int:
        return len(self.prompt_token_ids) + len(self.output_token_ids)

    def get_token_ids(self):
        return self.prompt_token_ids + self.output_token_ids

class SequenceGroup:
    def __init__(
        self,
        request_id: str,
        seqs: List[Sequence],
        arrival_time: float
    ):
        self.request_id = request_id

        if len(seqs) != len(set(seq.seq_id for seq in seqs)):
            raise ValueError("Duplicate seq_id found in sequences")

        self.seqs_dict = {seq.seq_id: seq for seq in seqs}
        self.arrival_time = arrival_time

    def get_seqs(self, status: Optional[SequenceStatus]=None) -> List[Sequence]:
        if not status:
            return list(self.seqs_dict.values())
        return [seq for seq in self.seqs_dict.values() if seq.status == status]

    def is_finished(self) -> bool:
        return all([seq.status == SequenceStatus.FINISHED for seq in self.seqs_dict.values()])


from core.block_manager import BlockAllocator, BlockSpaceManager
from core.scheduler import Scheduler
from core.sequence import Sequence, SequenceGroup


def test_simple_simulation_flow():
    block_size = 4
    num_blocks = 10
    allocator = BlockAllocator(num_blocks=num_blocks, block_size=block_size)
    block_manager = BlockSpaceManager(allocator)
    scheduler = Scheduler(block_manager)

    seq = Sequence(seq_id=1, prompt="Hello World", prompt_token_ids=[101, 102])
    seq_group = SequenceGroup(request_id="req_1", seqs=[seq], arrival_time=0.0)

    scheduler.add_sequence_group(seq_group)
    assert len(scheduler.waiting) == 1
    assert len(scheduler.running) == 0

    running_groups = scheduler.schedule()

    assert len(running_groups) == 1
    assert running_groups[0] == seq_group
    assert len(scheduler.waiting) == 0

    assert allocator.get_num_free_blocks() == 9
    assert len(block_manager.block_tables[1]) == 1

    seq.append_token_id(201, 1.0)
    block_manager.append_slot(seq.seq_id, seq.get_len() - 1)

    seq.append_token_id(202, 1.0)

    seq.append_token_id(203, 1.0)

    block_manager.append_slot(seq.seq_id, 4)

    assert allocator.get_num_free_blocks() == 8
    assert len(block_manager.block_tables[1]) == 2

    print("Simulation Test Passed!!!")

if __name__=="__main__":
    test_simple_simulation_flow()


import collections
import logging
from collections import OrderedDict
from dataclasses import dataclass
from typing import Deque, Dict, List

logger = logging.getLogger(__name__)

BLOCK_SIZE = 16

@dataclass
class PhysicalTokenBlock:
    device: str
    block_num: int
    block_size: int
    ref_count: int = 0

class BlockAllocator:
    def __init__(self, num_blocks: int, block_size: int=BLOCK_SIZE, device: str="cuda"):
        self.num_blocks = num_blocks
        self.block_size = block_size
        self.device = device

        self.free_blocks: Deque[PhysicalTokenBlock] = collections.deque()
        for i in range(num_blocks):
            self.free_blocks.append(PhysicalTokenBlock(device=device, block_num=i, block_size=block_size))

        self.all_blocks: Dict[int, PhysicalTokenBlock] = {
            b.block_num: b for b in self.free_blocks
        }

    def allocate(self) -> PhysicalTokenBlock:
        if not self.free_blocks:
            raise ValueError("Out of memory! No free blocks available")

        block = self.free_blocks.popleft()
        block.ref_count = 1
        return block

    def free(self, block: PhysicalTokenBlock):
        if block.ref_count == 0:
            raise ValueError(f"Block {block.block_num} is already free")

        block.ref_count -= 1
        if block.ref_count == 0:
            self.free_blocks.append(block)

    def get_num_free_blocks(self) -> int:
        return len(self.free_blocks)


class BlockSpaceManager:
    def __init__(self, block_allocator: BlockAllocator, prefix_cache: bool=True):
        self.allocator = block_allocator
        # map: seq_id --> list of PhysicalTokenBlock
        self.prefix_cache = prefix_cache
        self.block_tables: Dict[int, List[PhysicalTokenBlock]] = {}
        self.cached_blocks: OrderedDict[int, PhysicalTokenBlock] = OrderedDict()

    def _compute_block_hash(self, token_ids: List[int]) -> int:
        return hash(tuple(token_ids))

    def _is_block_cached(self, block: PhysicalTokenBlock) -> bool:
        return block in self.cached_blocks.values()

    def _evict_one_block(self) -> bool:
        for content_hash in list(self.cached_blocks.keys()):
            block = self.cached_blocks[content_hash]
            if block.ref_count == 0:
                del self.cached_blocks[content_hash]
                self.allocator.free_blocks.append(block)
                logger.info(f"EVICTED: block {block.block_num}")
                return True
        return False

    def allocate_with_prefix_cache(self, seq_id: int, token_ids: List[int]):
        if seq_id in self.block_tables:
            self.free(seq_id)

        blocks = []
        block_size = self.allocator.block_size

        for i in range(0, len(token_ids), block_size):
            chunk = token_ids[i:i+block_size]

            if len(chunk) == block_size:
                content_hash = self._compute_block_hash(chunk)

                if content_hash in self.cached_blocks:
                    logger.info(f"CACHE HIT: block {i//block_size} for seq {seq_id}")
                    block = self.cached_blocks[content_hash]
                    block.ref_count += 1
                    self.cached_blocks.move_to_end(content_hash)
                else:
                    logger.info(f"CACHE MISS: block {i//block_size} for seq {seq_id}")
                    if self.allocator.get_num_free_blocks() == 0:
                        if not self._evict_one_block():
                            raise RuntimeError("OOM: no blocks to evict")

                    block = self.allocator.allocate()
                    self.cached_blocks[content_hash] = block
            
                blocks.append(block)
            else:
                blocks.append(self.allocator.allocate())
        
        self.block_tables[seq_id] = blocks

    def allocate(self, seq_id: int, num_tokens: int):
        if seq_id in self.block_tables:
            self.free(seq_id)

        num_blocks = (num_tokens + self.allocator.block_size - 1) // self.allocator.block_size

        blocks = []
        try:
            for _ in range(num_blocks):
                blocks.append(self.allocator.allocate())
            self.block_tables[seq_id] = blocks
        except ValueError:
            for block in blocks:
                self.allocator.free(block)
            raise

    def allocate_request(self, seq_id: int, token_ids: List[int]):
        if self.prefix_cache:
            self.allocate_with_prefix_cache(seq_id, token_ids)
        else:
            self.allocate(seq_id, len(token_ids))

    def append_slot(self, seq_id: int, current_num_tokens: int):
        if seq_id not in self.block_tables:
            raise ValueError(f"Sequence {seq_id} not found in block tables")
        # if we have space in the previous block, do nothing
        if current_num_tokens % self.allocator.block_size != 0:
            return
        
        # allocate more blocks if there is no space
        self.block_tables[seq_id].append(self.allocator.allocate())

    def free(self, seq_id):
        if seq_id not in self.block_tables:
            return 

        # iterate and free the blocks in the block table
        for block in self.block_tables[seq_id]:
            if self._is_block_cached(block):
                block.ref_count -= 1
            else:
                self.allocator.free(block)

        # delete the entry from the dict as well
        del self.block_tables[seq_id]

    def can_allocate(self, num_tokens: int) -> bool:
        num_blocks_needed = (num_tokens + self.allocator.block_size - 1) // self.allocator.block_size
        return len(self.allocator.free_blocks) >= num_blocks_needed

    def can_allocate_with_cache(self, token_ids: List[int]) -> bool:
        block_size = self.allocator.block_size
        blocks_needed = 0

        for i in range(0, len(token_ids), block_size):
            chunk = token_ids[i:i+block_size]

            if len(chunk) == block_size:
                content_hash = self._compute_block_hash(chunk)
                if content_hash not in self.cached_blocks:
                    blocks_needed += 1
            else:
                blocks_needed += 1

        evictable_blocks = sum(1 for b in self.cached_blocks.values() if b.ref_count == 0)
        available_blocks = self.allocator.get_num_free_blocks() + evictable_blocks

        return available_blocks >= blocks_needed

    def can_allocate_request(self, token_ids: List[int]) -> bool:
        if self.prefix_cache:
            return self.can_allocate_with_cache(token_ids)
        else:
            return self.can_allocate(len(token_ids))

    def get_block_table(self, seq_id: int) -> List[int]:
        if seq_id not in self.block_tables:
            return []
        return [block.block_num for block in self.block_tables[seq_id]]
